@@ -1,12 +1,84 @@
-#' Store operations in a DelayedArray
+#' Store/reload a DelayedArray
 #'
-#' Store the delayed operations of a \linkS4class{DelayedArray} in a HDF5 file.
+#' Store or reload the delayed operations or array-like seeds of a \linkS4class{DelayedArray} in an existing HDF5 file.
 #'
-#' @param x Any of the delayed operation classes from \pkg{DelayedArray}.
-#' @param file String containing the path to a HDF5 file.
-#' @param name String containing the name of the group to save into.
-#' @param ... Arguments to be passed to specific methods.
-#' @return The contents of \code{x} are saved to \code{file}, and \code{NULL} is invisibly returned.
+#' @param x Any of the delayed operation/array classes from \pkg{DelayedArray}.
+#' @param handle An \pkg{rhdf5} handle of a HDF5 file to save into (for \code{storeDelayedObject}) or load from (for \code{reloadDelayedObject}).
+#' @param name String containing the name of the group in \code{file} to save into (for \code{storeDelayedObject}) or load from (for \code{reloadDelayedObject}).
+#' @param ... For \code{storeDelayedObject} and \code{reloadDelayedObject}, additional arguments to be passed to specific methods.
+#'
+#' For \code{altStoreDelayedObject} and \code{altReloadDelayedObject}, arguments to be passed to the alternative functions.
+#' @param version Package version of the \pkg{chihaya} format to use when storing or reloading delayed objects.
+#' When reloading, the version should be retrieved from the attributes of the outermost group, typically by \code{readDelayedArray}.
+#' @param reload Function to reload delayed objects from file.
+#' This should accept the same arguments as \code{reloadDelayedObject}.
+#' @param store Function (typically a generic) to store delayed objects to file.
+#' This should accept the same arguments as \code{storeDelayedObject}.
+#' @param type String specifying the type of delayed object, i.e., operation or array.
+#' This corresponds to \code{delayed_type} type in the \pkg{chihaya} attributes.
+#' @param subtype String specifying the subtype of the delayed object, 
+#' This corresponds to \code{delayed_array} or \code{delayed_operation} type (depending on \code{type}) in the \pkg{chihaya} attributes.
+#' @param fun Function to reload a delayed object.
+#' This should accept the same arguments as \code{reloadDelayedObject} and should return a delayed array (if \code{type="array"}) or operation (otherwise).
+#' It may also be \code{NULL} to delete an existing entry in the registry.
+#' @param existing Logical scalar indicating the action to take if a function has already been registered for \code{type} and \code{subtype} -
+#' keep the old or new function, or throw an error.
+#' @param save.external.array Logical scalar indicating whether to save an array-like seed as an external seed,
+#' even if a dedicated \code{storeDelayedObject} method is available.
+#' @param external.save.args Named list of further arguments to pass to \code{\link{altSaveObject}} when saving an external seed.
+#' @param external.dedup.session Session object created by \code{createExternalSeedDedupSession}.
+#' @param external.dedup.action String specifying the deduplication method to use.
+#'
+#' @section Customization:
+#' Developers can easily extend \pkg{alabaster.matrix} to new delayed objects by writing new methods for \code{storeDelayedObject}.
+#' Methods should save the contents of the delayed object to the HDF5 file in the \pkg{chihaya} format.
+#' Each new store method typically requires a corresponding reloading function to be registered via \code{registerReloadDelayedObjectFunction},
+#' so that \code{reloadDelayedObject} knows how to reconstitute the object from file.
+#' 
+#' Application developers can customize the process of storing/reloading delayed objects by specifying alternative functions in \code{altReloadDelayedObjectFunction} and \code{altStoreDelayedObjectFunction}.
+#' For example, if we want to preserve all delayed objects except for \linkS4class{DelayedSubset},
+#' we could replace \code{storeDelayedObject} with an \code{altStoreDelayedObject} that realizes any DelayedSubset instance into an ordinary matrix.
+#' This is analogous to the overrides for \code{\link{altReadObject}} and \code{\link{altSaveObject}}.
+#'
+#' Extension developers (i.e., those who write new methods for \code{storeDelayedObject} or new functions for \code{reloadDelayedObject})
+#' should generally use \code{altStoreDelayedObject} and \code{altReloadDelayedObject} in their method/funcion bodies.
+#' This ensures that any custom overrides specified by application developers are still respected in the extensions to \pkg{alabaster.matrix}.
+#'
+#' @section External seeds:
+#' Whenever \code{storeDelayedObject} encounters a delayed operation or array-like seed for which it has no methods,
+#' the ANY method will save the delayed object as an \dQuote{external seed}.
+#' The array is saved via \code{\link{altSaveObject}} into a \code{seeds} directory next to the file associated with \code{handle}.
+#' A reference to this external location is then stored in the \code{name} group inside \code{handle}.
+#'
+#' Users can force this behavior for all array-like seeds by passing \code{save.external.array=TRUE} in the \code{...} arguments of \code{storeDelayedObject}.
+#' This instructs \code{storeDelayedObject} to save everything as external seeds, including those arrays for which it has methods.
+#' Doing so can be beneficial to enable deduplication, e.g., when two delayed arrays perform different operations on the same underlying seed.
+#' By saving the seeds externally, file management systems can identify the redundancy to save storage space.
+#'
+#' Advanced users can explicitly deduplicate external seeds by setting \code{save.external.array=TRUE} and passing \code{external.dedup.session=} to \code{storeDelayedObject}.
+#' The \code{external.dedup.session} object is filled up with unique seeds as \code{storeDelayedObject} is called on various DelayedArrays.
+#' Whenever a duplicate seed is encountered, it is not saved again, but is instead linked or copied from the file path associated with the identical external seed.
+#' For example, a new session can be created when saving a SummarizedExperiment to deduplicate seeds across its assays.
+#'
+#' The exact deduplication action can be specified by specifying the \code{external.dedup.action=} parameter.
+#' By default, \code{storeDelayedObject} attempts to create hard links, falling back to copies when a link cannot be created.
+#' Users can instead create copies, symbolic links to absolute paths, or even symbolic links to relative paths
+#' (e.g., to link to a \dQuote{neighboring} assay of the same SummarizedExperiment).
+#'
+#' When external seeds are encountered by \code{reloadDelayedObject}, they are loaded as \linkS4class{ReloadedArray}s (or some variant thereof) by \code{\link{altReadObject}}.
+#' Users can forcibly realize the reloaded seed into memory by passing \code{custom.takane.reload=TRUE} in \code{...} for the \code{reloadDelayedObject} call.
+#' This is occasionally helpful for providing a more faithful roundtrip from file back into memory. 
+#' 
+#' @return 
+#' For \code{storeDelayedObject} and \code{altStoreDelayedObject}, the contents of \code{x} are saved to \code{file}, and \code{NULL} is invisibly returned.
+#'
+#' For \code{reloadDelayedObject} and \code{altReloadDelayedObject}, a delayed operation or \linkS4class{DelayedArray} is returned.
+#'
+#' For \code{altStoreDelayedObjectFunction}, the current store function is returned if \code{store} is missing.
+#' Otherwise, \code{store} is set as the current store function and the previous store function is returned.
+#'
+#' For \code{altReloadDelayedObjectFunction}, the current reload function is returned if \code{reload} is missing.
+#' Otherwise, \code{reload} is set as the current reload function and the previous reload function is returned.
 #' 
 #' @author Aaron Lun
 #' @examples
@@ -28,9 +100,12 @@
 #' reloadDelayedObject(fhandle, "YAY")
 #' H5Fclose(fhandle)
 #'
+#' @seealso
+#' \code{\link{saveObject,DelayedArray-method}} and \code{\link{readDelayedArray}}, where these methods are used.
+#'
+#' \url{https://artifactdb.github.io/chihaya/}, for the file format specification of delayed objects.
+#'
 #' @aliases
-#' storeDelayedObject
-#' reloadDelayedObject
 #' storeDelayedObject,ConstantArraySeed-method
 #' storeDelayedObject,DelayedAbind-method
 #' storeDelayedObject,ANY-method
@@ -108,11 +183,13 @@ save_vector_for_chihaya <- function(handle, name, x, version, scalar) {
 #######################################################
 #######################################################
 
-chihaya_array_registry <- list()
-chihaya_operation_registry <- list()
-chihaya_type_hint_registry <- list()
+chihaya.registry <- new.env()
+chihaya.registry$array <- list()
+chihaya.registry$operation <- list()
+chihaya.registry$type.hint <- list()
 
 #' @export
+#' @rdname storeDelayedObject
 reloadDelayedObject <- function(handle, name, version=package_version("1.1"), ...) {
     ghandle <- H5Gopen(handle, name)
     on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)
@@ -120,8 +197,8 @@ reloadDelayedObject <- function(handle, name, version=package_version("1.1"), ..
     output <- NULL
     if (h5_object_exists(ghandle, "_r_type_hint")) {
         hint <- h5_read_vector(ghandle, "_r_type_hint")
-        if (hint %in% names(chihaya_type_hint_registry)) {
-            FUN <- chihaya_type_hint_registry[[hint]]
+        if (hint %in% names(chihaya.registry$type.hint)) {
+            FUN <- chihaya.registry$type.hint[[hint]]
             output <- tryCatch(
                 FUN(ghandle, version=version, ...),
                 error=function(e) NULL
@@ -133,10 +210,10 @@ reloadDelayedObject <- function(handle, name, version=package_version("1.1"), ..
         objtype <- h5_read_attribute(ghandle, "delayed_type")
         if (objtype == "array") {
             arrtype <- h5_read_attribute(ghandle, "delayed_array")
-            FUN <- chihaya_array_registry[[arrtype]]
+            FUN <- chihaya.registry$array[[arrtype]]
         } else {
             optype <- h5_read_attribute(ghandle, "delayed_operation")
-            FUN <- chihaya_operation_registry[[optype]]
+            FUN <- chihaya.registry$operation[[optype]]
         }
 
         if (is.null(FUN)) {
@@ -151,6 +228,31 @@ reloadDelayedObject <- function(handle, name, version=package_version("1.1"), ..
     output
 }
 
+#' @export
+#' @rdname storeDelayedObject
+reloadDelayedObjectFunctionRegistry <- function(type=c("operation", "array")) {
+    type <- match.arg(type)
+    chihaya.registry[[type]]
+}
+
+#' @export
+#' @rdname storeDelayedObject
+registerReloadDelayedObjectFunction <- function(type=c("operation", "array"), subtype, fun, existing=c("old", "new", "error")) {
+    type <- match.arg(type)
+    if (!is.null(fun)) {
+        if (!is.null(chihaya.registry[[type]][[subtype]])) {
+            existing <- match.arg(existing)
+            if (existing == "old") {
+                return(invisible(NULL))
+            } else if (existing == "error") {
+                stop("function has already been registered for object type '", type, "'")
+            }
+        }
+    }
+    chihaya.registry[[type]][[subtype]] <- fun
+    invisible(NULL)
+}
+
 #######################################################
 #######################################################
 
@@ -167,7 +269,7 @@ setMethod("storeDelayedObject", "ConstantArraySeed", function(x, handle, name, v
 })
 
 #' @import rhdf5 DelayedArray
-chihaya_array_registry[["constant array"]] <- function(handle, version, ...) {
+chihaya.registry$array[["constant array"]] <- function(handle, version, ...) {
     dim <- h5_read_vector(handle, "dimensions")
     val <- load_vector_for_chihaya(handle, "value", version=version)
     ConstantArray(dim, value=val)
@@ -212,9 +314,10 @@ save_dense_array_for_chihaya <- function(x, handle, name, extract.native, versio
 }
 
 #' @export
+#' @rdname storeDelayedObject
 setMethod("storeDelayedObject", "array", function(x, handle, name, version=package_version("1.1"), save.external.array=FALSE, ...) {
     if (save.external.array) {
-        return(callNextMethod()) # calls the ANY method
+        return(selectMethod("storeDelayedObject", "ANY")(x, handle, name, version=version, save.external.array=save.external.array, ...))
     }
     save_dense_array_for_chihaya(x, handle, name, extract.native=identity, version=version, ...)
 })
@@ -222,7 +325,7 @@ setMethod("storeDelayedObject", "array", function(x, handle, name, version=packa
 #' @export
 setMethod("storeDelayedObject", "denseMatrix", function(x, handle, name, version=package_version("1.1"), save.external.array=FALSE, ...) {
     if (save.external.array) {
-        return(callNextMethod()) # calls the ANY method.
+        return(selectMethod("storeDelayedObject", "ANY")(x, handle, name, version=version, save.external.array=save.external.array, ...))
     } 
 
     extract.native <- NULL
@@ -233,7 +336,7 @@ setMethod("storeDelayedObject", "denseMatrix", function(x, handle, name, version
 })
 
 #' @import rhdf5 DelayedArray
-chihaya_array_registry[["dense array"]] <- function(handle, version, ...) {
+chihaya.registry$array[["dense array"]] <- function(handle, version, ...) {
     dhandle <- H5Dopen(handle, "data")
     on.exit(H5Dclose(dhandle), add=TRUE, after=FALSE)
     data <- H5Dread(dhandle)
@@ -297,7 +400,7 @@ save_sparse_matrix_for_chihaya <- function(x, handle, name, version=package_vers
 #' @export
 setMethod("storeDelayedObject", "sparseMatrix", function(x, handle, name, version=package_version("1.1"), save.external.array=FALSE, ...) {
     if (save.external.array) {
-        return(callNextMethod()) # calls the ANY method
+        return(selectMethod("storeDelayedObject", "ANY")(x, handle, name, version=version, save.external.array=save.external.array, ...))
     }
     save_sparse_matrix_for_chihaya(x, handle, name, version=version, ...)
 })
@@ -305,13 +408,13 @@ setMethod("storeDelayedObject", "sparseMatrix", function(x, handle, name, versio
 #' @export
 setMethod("storeDelayedObject", "SVT_SparseMatrix", function(x, handle, name, version=package_version("1.1"), save.external.array=FALSE, ...) {
     if (save.external.array) {
-        return(callNextMethod()) # calls the ANY method
+        return(selectMethod("storeDelayedObject", "ANY")(x, handle, name, version=version, save.external.array=save.external.array, ...))
     }
     save_sparse_matrix_for_chihaya(x, handle, name, version=version, ...)
 })
 
 #' @import rhdf5 DelayedArray
-chihaya_array_registry[["sparse matrix"]] <- function(handle, version, ...) {
+chihaya.registry$array[["sparse matrix"]] <- function(handle, version, ...) {
     indices <- h5_read_vector(handle, "indices")
     indptr <- h5_read_vector(handle, "indptr")
     data <- load_vector_for_chihaya(handle, "data", version=version)
@@ -320,7 +423,8 @@ chihaya_array_registry[["sparse matrix"]] <- function(handle, version, ...) {
     for (i in seq_along(svt)) {
         idx <- indptr[i] + seq_len(indptr[i+1] - indptr[i])
         if (length(idx)) {
-            svt[[i]] <- list(indices[idx], data[idx])
+            # As of version >= 1, data is first and indices are second.
+            svt[[i]] <- list(data[idx], indices[idx])
         }
     }
 
@@ -362,14 +466,14 @@ setMethod("storeDelayedObject", "DelayedAbind", function(x, handle, name, versio
     h5_write_attribute(shandle, "length", length(x@seeds), type="H5T_NATIVE_UINT32", scalar=TRUE)
 
     for (i in seq_along(x@seeds)) {
-        storeDelayedObject(x@seeds[[i]], shandle, as.character(i - 1L), version=version, ...)
+        altStoreDelayedObject(x@seeds[[i]], shandle, as.character(i - 1L), version=version, ...)
     }
 
     invisible(NULL)
 })
 
 #' @import alabaster.base rhdf5 DelayedArray
-chihaya_operation_registry[["combine"]] <- function(handle, version, ...) {
+chihaya.registry$operation[["combine"]] <- function(handle, version, ...) {
     shandle <- H5Gopen(handle, "seeds")
     on.exit(H5Gclose(shandle), add=TRUE, after=FALSE)
 
@@ -378,7 +482,7 @@ chihaya_operation_registry[["combine"]] <- function(handle, version, ...) {
 
     seeds <- vector("list", len)
     for (i in seq_len(len)) {
-        seeds[[i]] <- reloadDelayedObject(shandle, as.character(i - 1L), version=version, ...)
+        seeds[[i]] <- altReloadDelayedObject(shandle, as.character(i - 1L), version=version, ...)
     }
 
     if (along == 0L) {
@@ -401,13 +505,13 @@ setMethod("storeDelayedObject", "DelayedAperm", function(x, handle, name, versio
     h5_write_attribute(ghandle, "delayed_operation", "transpose", scalar=TRUE)
 
     h5_write_vector(ghandle, "permutation", x@perm - 1L, type="H5T_NATIVE_UINT32")
-    storeDelayedObject(x@seed, ghandle, "seed", version=version, ...)
+    altStoreDelayedObject(x@seed, ghandle, "seed", version=version, ...)
     invisible(NULL)
 })
 
 #' @import DelayedArray rhdf5
-chihaya_operation_registry[["transpose"]] <- function(handle, version, ...) {
-    x <- reloadDelayedObject(handle, "seed", version=version, ...)
+chihaya.registry$operation[["transpose"]] <- function(handle, version, ...) {
+    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
     perm <- h5_read_vector(handle, "permutation")
     aperm(x, perm + 1L)
 }
@@ -477,15 +581,15 @@ setMethod("storeDelayedObject", "DelayedNaryIsoOp", function(x, handle, name, ve
         stop("expected no additional right arguments for 'DelayedNaryIsoOp'")
     }
 
-    storeDelayedObject(x@seeds[[1]], ghandle, "left")
-    storeDelayedObject(x@seeds[[2]], ghandle, "right")
+    altStoreDelayedObject(x@seeds[[1]], ghandle, "left")
+    altStoreDelayedObject(x@seeds[[2]], ghandle, "right")
     invisible(NULL)
 })
 
 #' @import DelayedArray
 chihaya_load_binary_op <- function(handle, version, logic, ...) {
-    left <- reloadDelayedObject(handle, "left", version=version, ...)
-    right <- reloadDelayedObject(handle, "right", version=version, ...)
+    left <- altReloadDelayedObject(handle, "left", version=version, ...)
+    right <- altReloadDelayedObject(handle, "right", version=version, ...)
     op <- h5_read_vector(handle, "method")
     if (logic) {
         op <- translate_logic_Ops_from_chihaya(op)
@@ -493,9 +597,9 @@ chihaya_load_binary_op <- function(handle, version, logic, ...) {
     get(op, envir=baseenv())(left, right)
 }
 
-chihaya_operation_registry[["binary arithmetic"]] <- function(handle, version, ...) chihaya_load_binary_op(handle, version, logic=FALSE, ...)
-chihaya_operation_registry[["binary comparison"]] <- function(handle, version, ...) chihaya_load_binary_op(handle, version, logic=FALSE, ...)
-chihaya_operation_registry[["binary logic"]] <- function(handle, version, ...) chihaya_load_binary_op(handle=handle, version=version, logic=TRUE, ...)
+chihaya.registry$operation[["binary arithmetic"]] <- function(handle, version, ...) chihaya_load_binary_op(handle, version, logic=FALSE, ...)
+chihaya.registry$operation[["binary comparison"]] <- function(handle, version, ...) chihaya_load_binary_op(handle, version, logic=FALSE, ...)
+chihaya.registry$operation[["binary logic"]] <- function(handle, version, ...) chihaya_load_binary_op(handle=handle, version=version, logic=TRUE, ...)
 
 #######################################################
 #######################################################
@@ -541,13 +645,13 @@ setMethod("storeDelayedObject", "DelayedSetDimnames", function(x, handle, name,
     h5_write_attribute(ghandle, "delayed_operation", "dimnames", scalar=TRUE)
     save_dimnames_for_chihaya(ghandle, x@dimnames)
 
-    storeDelayedObject(x@seed, ghandle, "seed", version=version, ...)
+    altStoreDelayedObject(x@seed, ghandle, "seed", version=version, ...)
     invisible(NULL)
 })
 
 #' @import rhdf5 DelayedArray
-chihaya_operation_registry[["dimnames"]] <- function(handle, version, ...) {
-    x <- reloadDelayedObject(handle, "seed", version=version, ...)
+chihaya.registry$operation[["dimnames"]] <- function(handle, version, ...) {
+    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
     dimnames(x) <- load_dimnames_for_chihaya(handle)
     x
 }
@@ -598,15 +702,15 @@ setMethod("storeDelayedObject", "DelayedSubassign", function(x, handle, name, ve
     h5_write_attribute(ghandle, "delayed_operation", "subset assignment", scalar=TRUE)
 
     save_chihaya_indices(ghandle, "index", x@Lindex) 
-    storeDelayedObject(x@seed, ghandle, "seed", version=version, ...)
-    storeDelayedObject(x@Rvalue, ghandle, "value", version=version, ...)
+    altStoreDelayedObject(x@seed, ghandle, "seed", version=version, ...)
+    altStoreDelayedObject(x@Rvalue, ghandle, "value", version=version, ...)
     invisible(NULL)
 })
 
 #' @import rhdf5 DelayedArray
-chihaya_operation_registry[["subset assignment"]] <- function(handle, version, ...) {
-    x <- reloadDelayedObject(handle, "seed", version=version, ...)
-    value <- reloadDelayedObject(handle, "value", version=version, ...)
+chihaya.registry$operation[["subset assignment"]] <- function(handle, version, ...) {
+    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
+    value <- altReloadDelayedObject(handle, "value", version=version, ...)
     indices <- load_chihaya_indices(handle, "index")
     do.call(`[<-`, c(list(x=x), indices, list(value=value)))
 } 
@@ -624,13 +728,13 @@ setMethod("storeDelayedObject", "DelayedSubset", function(x, handle, name, versi
     h5_write_attribute(ghandle, "delayed_operation", "subset", scalar=TRUE)
 
     save_chihaya_indices(ghandle, "index", x@index) 
-    storeDelayedObject(x@seed, ghandle, "seed", version=version, ...)
+    altStoreDelayedObject(x@seed, ghandle, "seed", version=version, ...)
     invisible(NULL)
 })
 
 #' @import rhdf5 DelayedArray
-chihaya_operation_registry[["subset"]] <- function(handle, version, ...) {
-    x <- reloadDelayedObject(handle, "seed", version=version, ...)
+chihaya.registry$operation[["subset"]] <- function(handle, version, ...) {
+    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
     indices <- load_chihaya_indices(handle, "index")
     do.call(`[`, c(list(x), indices, list(drop=FALSE)))
 } 
@@ -764,7 +868,7 @@ chihaya.dump.unary.other <- function(handle, OP) {
 
 save_iso_op_stack <- function(handle, name, OPS, i, seed, version, ...) {
     if (i == 0L) {
-        storeDelayedObject(seed, handle, name, version=version, ...)
+        altStoreDelayedObject(seed, handle, name, version=version, ...)
         return(NULL)
     }
 
@@ -855,12 +959,12 @@ setMethod("storeDelayedObject", "DelayedUnaryIsoOpWithArgs", function(x, handle,
         stop("multi-dimensional 'value' not supported in 'DelayedUnaryIsoOpWithArgs'")
     }
 
-    storeDelayedObject(x@seed, ghandle, "seed", version=version, ...) 
+    altStoreDelayedObject(x@seed, ghandle, "seed", version=version, ...) 
     invisible(NULL)
 })
 
-chihaya_operation_registry[["unary math"]] <- function(handle, version, ...) {
-    x <- reloadDelayedObject(handle, "seed", version=version, ...)
+chihaya.registry$operation[["unary math"]] <- function(handle, version, ...) {
+    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
     method <- h5_read_vector(handle, "method")
     output <- NULL
 
@@ -927,8 +1031,8 @@ apply_unary_op_with_value <- function(x, op, side, handle, version) {
     output
 }
 
-chihaya_operation_registry[["unary logic"]] <- function(handle, version, ...) {
-    x <- reloadDelayedObject(handle, "seed", version=version, ...)
+chihaya.registry$operation[["unary logic"]] <- function(handle, version, ...) {
+    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
     method <- h5_read_vector(handle, "method")
 
     output <- NULL
@@ -943,22 +1047,22 @@ chihaya_operation_registry[["unary logic"]] <- function(handle, version, ...) {
     output
 }
 
-chihaya_operation_registry[["unary comparison"]] <- function(handle, version, ...) {
-    x <- reloadDelayedObject(handle, "seed", version=version, ...)
+chihaya.registry$operation[["unary comparison"]] <- function(handle, version, ...) {
+    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
     method <- h5_read_vector(handle, "method")
     side <- h5_read_vector(handle, "side")
     apply_unary_op_with_value(x, op=method, side=side, handle=handle, version=version)
 }
 
-chihaya_operation_registry[["unary special check"]] <- function(handle, version, ...) {
-    x <- reloadDelayedObject(handle, "seed", version=version, ...)
+chihaya.registry$operation[["unary special check"]] <- function(handle, version, ...) {
+    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
     method <- h5_read_vector(handle, "method")
     method <- sub("_", ".", method)
     get(method, envir=baseenv())(x)
 }
 
-chihaya_operation_registry[["unary arithmetic"]] <- function(handle, version, ...) {
-    x <- reloadDelayedObject(handle, "seed", version=version, ...)
+chihaya.registry$operation[["unary arithmetic"]] <- function(handle, version, ...) {
+    x <- altReloadDelayedObject(handle, "seed", version=version, ...)
     method <- h5_read_vector(handle, "method")
     side <- h5_read_vector(handle, "side")
 
@@ -983,7 +1087,17 @@ chihaya_operation_registry[["unary arithmetic"]] <- function(handle, version, ..
 
 #' @export
 #' @import rhdf5
-setMethod("storeDelayedObject", "ANY", function(x, handle, name, version=package_version("1.1"), ...) {
+#' @rdname storeDelayedObject
+setMethod("storeDelayedObject", "ANY", function(
+    x,
+    handle,
+    name,
+    version=package_version("1.1"),
+    external.save.args=list(),
+    external.dedup.session=NULL,
+    external.dedup.action=c("link", "copy", "symlink", "relsymlink"), 
+    ...) 
+{
     ghandle <- H5Gcreate(handle, name)
     on.exit(H5Gclose(ghandle), add=TRUE, after=FALSE)
 
@@ -991,9 +1105,9 @@ setMethod("storeDelayedObject", "ANY", function(x, handle, name, version=package
         h5_write_attribute(ghandle, "delayed_type", "operation", scalar=TRUE)
         h5_write_attribute(ghandle, "delayed_operation", "matrix product", scalar=TRUE)
 
-        storeDelayedObject(x@rotation, ghandle, "left_seed", version=version, ...)
+        altStoreDelayedObject(x@rotation, ghandle, "left_seed", version=version, ...)
         h5_write_vector(ghandle, "left_orientation", "N", scalar=TRUE)
-        storeDelayedObject(x@components, ghandle, "right_seed", version=version, ...)
+        altStoreDelayedObject(x@components, ghandle, "right_seed", version=version, ...)
         h5_write_vector(ghandle, "right_orientation", "T", scalar=TRUE)
 
     } else if (is(x, "ResidualMatrixSeed")) {
@@ -1013,16 +1127,16 @@ setMethod("storeDelayedObject", "ANY", function(x, handle, name, version=package
         # Mimic a binary subtraction.
         h5_write_attribute(xhandle, "delayed_operation", "binary arithmetic", scalar=TRUE)
         h5_write_vector(xhandle, "method", "-", scalar=TRUE)
-        storeDelayedObject(x@.matrix, xhandle, "left", version=version, ...)
+        altStoreDelayedObject(x@.matrix, xhandle, "left", version=version, ...)
 
         # Mimic a matrix product.
         rhandle <- H5Gcreate(xhandle, "right")
         on.exit(H5Gclose(rhandle), add=TRUE, after=FALSE)
         h5_write_attribute(rhandle, "delayed_type", "operation", scalar=TRUE)
         h5_write_attribute(rhandle, "delayed_operation", "matrix product", scalar=TRUE)
-        storeDelayedObject(x@Q, rhandle, "left_seed", version=version, ...)
+        altStoreDelayedObject(x@Q, rhandle, "left_seed", version=version, ...)
         h5_write_vector(rhandle, "left_orientation", "N", scalar=TRUE)
-        storeDelayedObject(x@Qty, rhandle, "right_seed", version=version, ...)
+        altStoreDelayedObject(x@Qty, rhandle, "right_seed", version=version, ...)
         h5_write_vector(rhandle, "right_orientation", "N", scalar=TRUE)
 
     } else {
@@ -1032,7 +1146,26 @@ setMethod("storeDelayedObject", "ANY", function(x, handle, name, version=package
         exdir <- file.path(dirname(H5Fget_name(handle)), "seeds")
         dir.create(exdir, showWarnings=FALSE)
         n <- length(list.files(exdir))
-        saveObject(x, file.path(exdir, n), ...)
+        output <- file.path(exdir, n)
+
+        dedup.path <- check_external_seed_in_dedup_session(x, session=external.dedup.session)
+        if (is.null(dedup.path)) {
+            # We try twice, once for the seed itself, and then again after
+            # wrapping it in a DelayedArray, just in case the seed has an Array
+            # class with an altSaveObject method. If not, the DelayedArray call
+            # will just fall back to the default dense/sparse savers.
+            exargs <- c(list(x=x, path=output), external.save.args)
+            status <- do.call(try_altSaveObject, exargs)
+            if (!status) {
+                exargs$x <- DelayedArray(exargs$x)
+                exargs$DelayedArray.dispatch.pristine <- FALSE # we only got here if we failed to save the seed itself, so no point trying again.
+                exargs$DelayedArray.preserve.ops <- FALSE # avoid infinite recursion between storeDelayedObject and saveObject.
+                do.call(altSaveObject, exargs)
+            }
+        } else {
+            clone_duplicate(dedup.path, output, action=match.arg(external.dedup.action))
+        }
+        add_external_seed_to_dedup_session(x, session=external.dedup.session, path=output)
 
         h5_write_vector(ghandle, "dimensions", dim(x), type="H5T_NATIVE_UINT32", compress=0)
         h5_write_vector(ghandle, "type", to_value_type(type(x)), scalar=TRUE)
@@ -1041,9 +1174,9 @@ setMethod("storeDelayedObject", "ANY", function(x, handle, name, version=package
 })
 
 #' @import alabaster.base rhdf5 DelayedArray
-chihaya_array_registry[["custom takane seed array"]] <- function(handle, version, custom.takane.realize=FALSE, ...) {
+chihaya.registry$array[["custom takane seed array"]] <- function(handle, version, custom.takane.realize=FALSE, ...) {
     index <- h5_read_vector(handle, "index")
-    out <- readObject(file.path(dirname(H5Fget_name(handle)), "seeds", index), ...)
+    out <- altReadObject(file.path(dirname(H5Fget_name(handle)), "seeds", index), ...)
 
     if (custom.takane.realize) {
         if (is_sparse(out)) {
@@ -1057,14 +1190,14 @@ chihaya_array_registry[["custom takane seed array"]] <- function(handle, version
 }
 
 #' @importFrom Matrix t
-chihaya_operation_registry[["matrix product"]] <- function(handle, version, ...) {
-    L <- as.matrix(reloadDelayedObject(handle, "left_seed"))
+chihaya.registry$operation[["matrix product"]] <- function(handle, version, ...) {
+    L <- as.matrix(altReloadDelayedObject(handle, "left_seed"))
     Lori <- h5_read_vector(handle, "left_orientation")
     if (length(Lori) == 1 && as.character(Lori) == "T") {
         L <- t(L)
     } 
 
-    R <- as.matrix(reloadDelayedObject(handle, "right_seed"))
+    R <- as.matrix(altReloadDelayedObject(handle, "right_seed"))
     Rori <- h5_read_vector(handle, "right_orientation")
     if (length(Rori) == 1 && as.character(Rori) == "N") {
         R <- t(R)
@@ -1073,7 +1206,7 @@ chihaya_operation_registry[["matrix product"]] <- function(handle, version, ...)
     BiocSingular::LowRankMatrix(L, R)
 }
 
-chihaya_type_hint_registry[["residual matrix"]] <- function(handle, version, ...) {
+chihaya.registry$type.hint[["residual matrix"]] <- function(handle, version, ...) {
     if (!isNamespaceLoaded("ResidualMatrix")) {
         loadNamespace("ResidualMatrix")
     }
@@ -1093,7 +1226,7 @@ chihaya_type_hint_registry[["residual matrix"]] <- function(handle, version, ...
 
     stopifnot(identical(optype, "binary arithmetic"))
     stopifnot(identical(h5_read_vector(xhandle, "method"), "-"))
-    .matrix <- reloadDelayedObject(xhandle, "left")
+    .matrix <- altReloadDelayedObject(xhandle, "left")
     .matrix <- .matrix@seed # can't be a DelayedArray inside ResidualMatrix, for various reasons...
 
     rhandle <- H5Gopen(xhandle, "right")
@@ -1103,8 +1236,103 @@ chihaya_type_hint_registry[["residual matrix"]] <- function(handle, version, ...
     stopifnot(identical(h5_read_vector(rhandle, "left_orientation"), "N"))
     stopifnot(identical(h5_read_vector(rhandle, "right_orientation"), "N"))
 
-    Q <- as.matrix(reloadDelayedObject(rhandle, "left_seed", version=version, ...))
-    Qty <- as.matrix(reloadDelayedObject(rhandle, "right_seed", version=version, ...))
+    Q <- as.matrix(altReloadDelayedObject(rhandle, "left_seed", version=version, ...))
+    Qty <- as.matrix(altReloadDelayedObject(rhandle, "right_seed", version=version, ...))
     seed <- new("ResidualMatrixSeed", .matrix = .matrix, Q = Q, Qty = Qty, transposed = transposed)
     DelayedArray(seed)
 }
+
+#######################################################
+#######################################################
+
+#' @export
+#' @rdname storeDelayedObject
+altStoreDelayedObjectFunction <- (function() {
+    cur.env <- new.env()
+    cur.env$store <- NULL
+    function(store) {
+        prev <- cur.env$store
+        if (missing(store)) {
+            prev
+        } else {
+            cur.env$store <- store
+            invisible(prev)
+        }
+    }
+})()
+
+#' @export
+#' @rdname storeDelayedObject
+altStoreDelayedObject <- function(...) {
+    FUN <- altStoreDelayedObjectFunction()
+    if (is.null(FUN)) {
+        FUN <- storeDelayedObject
+    }
+    FUN(...)
+}
+
+#' @export
+#' @rdname storeDelayedObject
+altReloadDelayedObjectFunction <- (function() {
+    cur.env <- new.env()
+    cur.env$reload <- NULL
+    function(reload) {
+        prev <- cur.env$reload
+        if (missing(reload)) {
+            prev
+        } else {
+            cur.env$reload <- reload
+            invisible(prev)
+        }
+    }
+})()
+
+#' @export
+#' @rdname storeDelayedObject
+altReloadDelayedObject <- function(...) {
+    FUN <- altReloadDelayedObjectFunction()
+    if (is.null(FUN)) {
+        FUN <- reloadDelayedObject
+    }
+    FUN(...)
+}
+
+#######################################################
+#######################################################
+
+#' @export
+#' @rdname storeDelayedObject
+createExternalSeedDedupSession <- function() {
+    output <- new.env()
+    output$known <- list()
+    output
+}
+
+check_external_seed_in_dedup_session <- function(x, session) {
+    if (is.null(session)) {
+        return(NULL)
+    }
+
+    cls <- as.character(class(x))[1]
+    if (!(cls %in% names(session$known))) {
+        return(NULL)
+    }
+
+    candidates <- session$known[[cls]]
+    for (y in candidates) {
+        if (identical(x, y$value)) {
+            return(y$path)
+        }
+    }
+
+    return(NULL)
+}
+
+add_external_seed_to_dedup_session <- function(x, session, path) {
+    cls <- as.character(class(x))[1]
+    if (!(cls %in% names(session$known))) {
+        session$known[[cls]] <- list()
+    }
+    path <- normalizePath(path, mustWork=TRUE)
+    session$known[[cls]] <- c(session$known[[cls]], list(list(value=x, path=path)))
+}
