@@ -181,7 +181,8 @@ test_that("sparse matrix saving handles dimnames", {
 
     temp <- saveDelayed(DelayedArray(X))
     roundtrip <- loadDelayed(temp)
-    expect_identical(as(X, "SVT_SparseMatrix"), roundtrip@seed)
+    expect_identical(type(X), type(roundtrip@seed))
+    expect_identical(X, as(roundtrip@seed, "dgCMatrix"))
 })
 
 test_that("sparse matrices are saved to external arrays if requested", {
@@ -727,3 +728,118 @@ test_that("saving of a ResidualMatrix works correctly", {
     expect_false(is(out, "ResidualMatrix"))
     expect_identical(unname(as.matrix(thing2)), unname(as.matrix(out)))
 })
+
+#######################################################
+#######################################################
+
+test_that("readDelayedObject registry works as expected", {
+    expect_false(is.null(reloadDelayedObjectFunctionRegistry("array")[["constant array"]])) 
+
+    expect_null(reloadDelayedObjectFunctionRegistry("array")[["foobar"]]) 
+    registerReloadDelayedObjectFunction("array", "foobar", function(...) "WHEE")
+    expect_identical(reloadDelayedObjectFunctionRegistry("array")[["foobar"]](), "WHEE")
+
+    expect_error(registerReloadDelayedObjectFunction("array", "foobar", function(...) "WHEE", "error"), "already been registered")
+    registerReloadDelayedObjectFunction("array", "foobar", function(...) "FART", "old")
+    expect_identical(reloadDelayedObjectFunctionRegistry("array")[["foobar"]](), "WHEE")
+    registerReloadDelayedObjectFunction("array", "foobar", function(...) "FART", "new")
+    expect_identical(reloadDelayedObjectFunctionRegistry("array")[["foobar"]](), "FART")
+
+    registerReloadDelayedObjectFunction("array", "foobar", NULL)
+    expect_null(reloadDelayedObjectFunctionRegistry("array")[["foobar"]]) 
+})
+
+test_that("alternative delayed object function overrides work", {
+    old <- altStoreDelayedObjectFunction()
+    altStoreDelayedObjectFunction(function() "whee")
+    on.exit(altStoreDelayedObjectFunction(old), after=FALSE, add=TRUE)
+    expect_identical(altStoreDelayedObject(), "whee")
+
+    old <- altReloadDelayedObjectFunction()
+    altReloadDelayedObjectFunction(function() "whee")
+    on.exit(altReloadDelayedObjectFunction(old), after=FALSE, add=TRUE)
+    expect_identical(altReloadDelayedObject(), "whee")
+})
+
+#######################################################
+#######################################################
+
+test_that("external saving handles seeds without saveObject methods", {
+    X <- DelayedArray(new("SuperSeed", dim=c(100L, 50L))) + 100
+
+    temp <- saveDelayed(X)
+    expect_true(file.exists(file.path(temp, "seeds", 0)))
+    roundtrip <- loadDelayed(temp, custom.takane.realize=TRUE)
+    expect_identical(as.matrix(X), as.matrix(roundtrip))
+})
+
+test_that("external deduplication is done correctly", {
+    dedup.session <- createExternalSeedDedupSession()
+
+    X <- DelayedArray(matrix(rpois(60, 5), ncol=20))
+    temp <- saveDelayed(X, save.external.array=TRUE, external.dedup.session=dedup.session)
+    expect_true(file.exists(file.path(temp, "seeds", 0)))
+    roundtrip <- loadDelayed(temp, custom.takane.realize=TRUE)
+    expect_identical(X, roundtrip)
+
+    Y <- X + 1
+    temp2a <- saveDelayed(Y, save.external.array=TRUE, external.dedup.session=dedup.session, external.dedup.action="link")
+    expect_true(file.exists(file.path(temp2a, "seeds", 0)))
+    roundtrip <- loadDelayed(temp2a, custom.takane.realize=TRUE)
+    expect_equal(Y, roundtrip)
+
+    if (.Platform$OS.type=="unix") { 
+        temp2b <- saveDelayed(Y, save.external.array=TRUE, external.dedup.session=dedup.session, external.dedup.action="symlink")
+        expect_true(file.exists(file.path(temp2b, "seeds", 0)))
+        roundtrip <- loadDelayed(temp2b, custom.takane.realize=TRUE)
+        expect_equal(Y, roundtrip)
+    }
+
+    temp2c <- saveDelayed(Y, save.external.array=TRUE, external.dedup.session=dedup.session, external.dedup.action="copy")
+    expect_true(file.exists(file.path(temp2c, "seeds", 0)))
+    roundtrip <- loadDelayed(temp2c, custom.takane.realize=TRUE)
+    expect_equal(Y, roundtrip)
+
+    Z <- DelayedArray(matrix(rpois(30, 5), ncol=5)) # checking that a different array doesn't trigger the deduplicator.
+    temp3 <- saveDelayed(Z, save.external.array=TRUE, external.dedup.session=dedup.session)
+    expect_true(file.exists(file.path(temp3, "seeds", 0)))
+    roundtrip <- loadDelayed(temp3, custom.takane.realize=TRUE)
+    expect_identical(Z, roundtrip)
+
+    if (.Platform$OS.type=="unix") { 
+        expect_identical(Sys.readlink(file.path(temp, "seeds", "0", "OBJECT")), "")
+        expect_identical(Sys.readlink(file.path(temp2a, "seeds", "0", "OBJECT")), "")
+        expect_true(startsWith(Sys.readlink(file.path(temp2b, "seeds", "0", "OBJECT")), "/"))
+        expect_identical(Sys.readlink(file.path(temp2c, "seeds", "0", "OBJECT")), "")
+        expect_identical(Sys.readlink(file.path(temp3, "seeds", "0", "OBJECT")), "")
+    }
+})
+
+test_that("external deduplication works with relative paths", {
+    skip_on_os("windows")
+    dedup.session <- createExternalSeedDedupSession()
+
+    staging <- tempfile()
+    dir.create(staging)
+    pwd <- getwd()
+    on.exit(setwd(pwd), after=FALSE, add=TRUE)
+    setwd(staging)
+
+    X <- DelayedArray(matrix(rpois(60, 5), ncol=20))
+    dir.create("original")
+    saveObject(X + 2, file.path("original", "out"), DelayedArray.preserve.ops=TRUE, 
+        DelayedArray.store.args=list(save.external.array=TRUE, external.dedup.session=dedup.session, external.dedup.action="relsymlink"))
+
+    Y <- X + 1
+    saveObject(Y, "semiclone", DelayedArray.preserve.ops=TRUE, 
+        DelayedArray.store.args=list(save.external.array=TRUE, external.dedup.session=dedup.session, external.dedup.action="relsymlink"))
+
+    # Changing back and seeing whether the loader works.
+    setwd(pwd)
+    expect_true(file.exists(file.path(staging, "semiclone", "seeds", 0)))
+    roundtrip <- loadDelayed(file.path(staging, "semiclone"), custom.takane.realize=TRUE)
+    expect_equal(Y, roundtrip)
+
+    expect_identical(Sys.readlink(file.path(staging, "original", "out", "seeds", "0", "OBJECT")), "")
+    expect_true(startsWith(Sys.readlink(file.path(staging, "semiclone", "seeds", "0", "OBJECT")), "../"))
+})
